// ***************************************************************************
// Name:        3D-Wireframe (class declaration file)
// Purpose:     shows wireframe 3d objects
// Copyright:   (C) 2003, 2004, 2005, 2007 Dominik Wagenfuehr <dominik.wagenfuehr@arcor.de>
// ***************************************************************************

/**
This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "objects.h"

// **************************
// Constants
// **************************

static const double PI=3.1415927;    // approx. pi                                         

// **************************
// Class Declaration
// **************************

static double pow(double a, int b);          // potency for double
static void message(wxString string, wxString title="");    // messagebox for internal use only
static void message(long int number, wxString title="");        // messagebox for internal use only
static void message(int number, wxString title="");        // messagebox for internal use only
static void message(double number, wxString title="");         // messagebox for internal use only

// *****************************
// Event-Handler
// *****************************

BEGIN_EVENT_TABLE(ImageFrame,wxFrame)
  EVT_PAINT(ImageFrame::OnPaint)
  EVT_MOUSE_EVENTS(ImageFrame::Look)
END_EVENT_TABLE()

// **************************
// Class Definitions
// **************************

double pow(double a, int b) {          // potency for double
  double prod=1;
  for (int i=0; i<b;i++) prod*=a;
  return(prod);
};

static void message(wxString string, wxString title) {        // messagebox for internal use only
  wxMessageBox( string, "Report "+ title, wxOK );
};

static void message(int number, wxString title) {        // messagebox for internal use only
  wxString string;
  string << number;
  wxMessageBox( string, "Report "+ title, wxOK );
};

static void message(long int number, wxString title) {        // messagebox for internal use only
  wxString string;
  string << number;
  wxMessageBox( string, "Report "+ title, wxOK );
};

static void message(double number, wxString title) {        // messagebox for internal use only
  wxString string;
  string << number;
  wxMessageBox( string, "Report "+ title, wxOK );
};

// **************************
// Vector Functions
// **************************

Vector::Vector(double x, double y, double z) {	  	  // standard contructor                                       
  xcoord=x;
  ycoord=y;
  zcoord=z;
};

Vector::Vector(void) {	  	                   // standard contructor                                       
  xcoord=0;
  ycoord=0;
  zcoord=0;
};


Vector Vector::operator=(Vector v) {	  	  // copy constructor                                       
  xcoord=v.x();
  ycoord=v.y();
  zcoord=v.z();
  return (*this);
};

Vector Vector::operator-(void) {	  	  // negation operator                                       
  return( Vector( -x(), -y(), -z() ) );
};

bool Vector::operator==(Vector v) {	  	  // compare operator                                      
  if (xcoord==v.x() && ycoord==v.y() && zcoord==v.z()) return (true);
  else return (false);
};

bool Vector::operator!=(Vector v) {	  	  // compare operator                                      
  if (xcoord==v.x() && ycoord==v.y() && zcoord==v.z()) return (false);
  else return (true);
};

Vector Vector::operator+(Vector v) {	  	  // binary addition operator                                       
  return( Vector( x()+v.x(), y()+v.y(), z()+v.z() ) );
};

Vector Vector::operator+=(Vector v) {	  	  // unary addition operator                                       
  xcoord+=v.x();
  ycoord+=v.y();
  zcoord+=v.z();
  return (*this);
};

Vector Vector::operator-(Vector v) {	  	  // binary subtraction operator                                       
  return( Vector( x()-v.x(), y()-v.y(), z()-v.z() ) );
};

Vector Vector::operator-=(Vector v) {	  	  // unary subtraction operator                                       
  xcoord-=v.x();
  ycoord-=v.y();
  zcoord-=v.z();
  return (*this);
};

Vector operator*(double r, Vector v) { 	        // binary scalar multiplication operator                                       
  return( Vector( r*v.x(), r*v.y(), r*v.z() ) );
};

Vector operator*(Vector v, double r) { 	       // binary scalar multiplication operator                                       
  return( Vector( r*v.x(), r*v.y(), r*v.z() ) );
};

Vector Vector::operator*=(double r) {	  	  // unary scalar multiplication operator                                       
  xcoord*=r;
  ycoord*=r;
  zcoord*=r;
  return (*this);
};

Vector Vector::VecProd(Vector v) const {          // cross product between two vectors
  return ( Vector( y()*v.z()-z()*v.y(), z()*v.x()-x()*v.z(), x()*v.y()-y()*v.x() ) );
};

double Vector::ScalProd(Vector v) const {           // scalar product between two vectors
  return(x()*v.x()+y()*v.y()+z()*v.z());
};

double Vector::Norm(void) const {                      // norm / length of a vector
  return(sqrt(ScalProd(*this)));
};

Vector Vector::Normalize(double norm) {             // normalize a vector to length n
  if (Norm()==0) return(*this);       // Nullvector
  *this*=(norm/Norm());
  return (*this);
};

Vector Vector::Orth(Vector r, double norm) const {        // calculates a vector that is orthogonal with norm
  Vector v=*this;
  v=(v.VecProd(r)).Normalize(norm);
  return(v);
};

Vector Vector::Translate(Vector v) {             // move a vector
  *this+=v;
  return (*this);
};

Vector Vector::Rotate(double angle, Vector r) {  // rotate the vector around vector axis r
  double c,s,t,theta;
  Vector w=r;

  if (w==Vector()) return(*this);
  w.Normalize();
  
  theta=angle*PI/180;                // Rotation angle in radian
  c=cos(theta); s=sin(theta); t=1-cos(theta);
  
  *this=Vector(
  (t*pow(w.x(),2)+c)*x()+(t*w.x()*w.y()-s*w.z())*y()+(t*w.x()*w.z()+s*w.y())*z(),
  (t*w.x()*w.y()+s*w.z())*x()+(t*pow(w.y(),2)+c)*y()+(t*w.y()*w.z()-s*w.x())*z(),
  (t*w.x()*w.z()-s*w.y())*x()+(t*w.y()*w.z()+s*w.x())*y()+(t*pow(w.z(),2)+c)*z());      // new coordiates

  return(*this);
};                                                         

double Vector::Angle(Vector v) const {                     // calculate the angle between two vectors
  double cosinus;
  Vector w=*this;
  if (v==Vector() || w==Vector()) return (0); 	// One vector is the null vector.
  else {
    cosinus=w.ScalProd(v)/(w.Norm()*v.Norm());
    if (cosinus>=1) cosinus=1;
    else if (cosinus<=-1) cosinus=-1;
    return(180*acos(cosinus)/PI);
  }     
};

double Vector::x(void) const {  // returns x-coord
  return(xcoord);
};

double Vector::y(void) const {  // returns y-coord
  return(ycoord);
};

double Vector::z(void) const {  // returns z-coord
  return(zcoord);
};

void Vector::Print(wxString string) const {           // prints all coordiantes on screen
  wxString str;
  str << "X: " << xcoord << "\nY: " << ycoord << "\nZ: " << zcoord;
  message (str, string);
};

// **************************
// ArrayOfVectors Functions
// **************************

ArrayOfVectors::ArrayOfVectors(void) {        // standard constructor
  anchor=NULL;
  size=0;
};

ArrayOfVectors::ArrayOfVectors(Vector v) {        // standard constructor
  anchor=NULL;
  size=0;
  Add(v);
};

/*
ArrayOfVectors::~ArrayOfVectors() {               // destructor
  if (IsEmpty()) return;         // list is empty

  VectorList *pointer=anchor;
  while (pointer->next) pointer=pointer->next;         // set pointer to end

  do {
    pointer=pointer->last;              // go backwards through the list
    delete (pointer->next);            // delete the elements ...
  } while (pointer->last);              // until you are at the anchor
  
  delete (anchor);                   // delete anchor
};
*/

Vector& ArrayOfVectors::operator[](size_t index) {
  return(GetItem(index));
};

void ArrayOfVectors::Add(Vector v) {                         // add a vector to the list
  if (IsEmpty()) {              // list is empty  
    anchor=new VectorList;
    anchor->last=NULL;
    anchor->next=NULL;
    anchor->element=v;  
  }
  else {
    VectorList *pointer=anchor; 
    while (pointer->next) pointer=pointer->next;        // set pointer to end
    pointer->next=new VectorList;
    (pointer->next)->last=pointer;
    pointer=pointer->next;
    pointer->next=NULL;
    pointer->element=v;
  }
  size++;
};

size_t ArrayOfVectors::Count(void) {                // count the elements in the array
  if (IsEmpty()) size=0;         // list is empty
  else {
    VectorList *pointer=anchor;    // pointer for going through the list
    size=1;                        // counter
    while (pointer->next) {  pointer=pointer->next; size++; }   // count elements
  }  
  return(size);
};

void ArrayOfVectors::Delete(size_t index) {                  // delete vector at position index
  if (index<=0 || index>size) return;          // index out of boundaries

  VectorList *pointer=anchor;  // set pointer to start
  
  if(index==1) {      // anchor should be deleted
    if (size==1) anchor=NULL;  // delete anchor
    else {
      anchor=pointer->next;   // set second element as anchor
      anchor->last=NULL;
    }
  }
  else if (index==size) {    // last element should be deleted
    while (pointer->next) pointer=pointer->next;        // set pointer to end
    (pointer->last)->next=NULL;
  }
  else {
    for (unsigned int i=1; i<index; i++) pointer=pointer->next;         // go to right element
    (pointer->last)->next=pointer->next;
    (pointer->next)->last=pointer->last;
  }
  delete (pointer);                   // delete element
  size--;
};

void ArrayOfVectors::Empty(void) {                           // clears list
  if (IsEmpty()) return;              // list is empty

  VectorList *pointer=anchor;         // set pointer to end
  while (pointer->next) pointer=pointer->next;        // set pointer to end
  do {
    pointer=pointer->last;              // go backwards trhough the list
    delete (pointer->next);           // delete the elements ...
  } while (pointer->last);              // until you are at the anchor
  
  delete (anchor);                   // delete anchor
  anchor=NULL;
  size=0;
};

VectorList *ArrayOfVectors::GetAnchor(void) const {          // get anchor of the array
  return(anchor);
};

size_t ArrayOfVectors::GetCount(void) const {                // get size of the array
  return(size);
};

Vector &ArrayOfVectors::GetItem(size_t index) {         // get item at position index
  if (index<=0 || index>size) {          // index out of boundaries
    if (!IsEmpty()) return(anchor->element);   // return first element, if there is one (else undefined)
  }  
  VectorList *pointer=anchor;           // set pointer to start
  for (unsigned int i=1; i<index; i++) pointer=pointer->next;         // go to right element
  return(pointer->element);                     // returns value
};

bool ArrayOfVectors::IsEmpty(void) const {                   // check if list is empty
  if (anchor) return(false);
  else return(true);
};

void ArrayOfVectors::SetItem(size_t index, Vector v) {       // set item at position index to vector
  if (index>0 && index<=size) {          // index out of boundaries
    VectorList *pointer=anchor;           // set pointer to start
    for (unsigned int i=1; i<index; i++) pointer=pointer->next;         // go to right element
    pointer->element=v;
  }
};

// **************************
// Vertex Functions
// **************************

Vertex::Vertex(double x, double y, double z) {               // standard contructor
  vertex=Vector(x,y,z);
  edge.Empty();
};

Vertex::Vertex(Vector v) {                                    // standard contructor
  vertex=v;
  edge.Empty();
};

/*
Vertex::~Vertex() {               // destructor
  puts("V1");
  edge.Empty();
  puts("V2");
};
*/

void Vertex::AddEdge(Vector v) {                         // add a vector to the list
  edge.Add(v);
};

void Vertex::DeleteEdge(size_t index) {                  // delete vector at position index
  edge.Delete(index);
};

void Vertex::EmptyEdges(void) {                           // clears list
  edge.Empty();
};

VectorList *Vertex::GetEdgeAnchor(void) const {          // get anchor of the array
  return(edge.GetAnchor());
};

size_t Vertex::GetEdgeCount(void) const {                // get size of the array
  return(edge.GetCount());
};

Vector &Vertex::GetEdgeItem(size_t index) {         // get item at position index
  return(edge.GetItem(index));
};

bool Vertex::IsEdgeEmpty(void) const {                   // check if list is empty
  return(edge.IsEmpty());
};

void Vertex::SetEdgeItem(size_t index, Vector v) {       // set item at position index to vector
  edge.SetItem(index, v);
};

Vertex Vertex::Rotate(double angle, Vector r) {  // rotate the vector and all edges around vector axis r
  for (unsigned int i=1; i<=edge.GetCount(); i++) {
    edge.SetItem(i, edge[i].Rotate(angle,r));
  }
  vertex.Rotate(angle,r);
  return(*this);
};

Vertex Vertex::Translate(Vector v) {             // move a vector and edges
  for (unsigned int i=1; i<=edge.GetCount(); i++) {
    edge.SetItem(i, edge[i].Translate(v));
  }
  vertex.Translate(v);
  return(*this);
};

Vector Vertex::GetVector(void) const {                        // returns vector
  return(vertex);
};

double Vertex::x(void) const {  // returns x-coord
  return(vertex.x());
};

double Vertex::y(void) const {  // returns y-coord
  return(vertex.y());
};

double Vertex::z(void) const {  // returns z-coord
  return(vertex.z());
};

// **************************
// ArrayOfVertices Functions
// **************************

ArrayOfVertices::ArrayOfVertices(void) {        // standard constructor
  anchor=NULL;
  size=0;
};

ArrayOfVertices::ArrayOfVertices(Vertex v) {        // standard constructor
  anchor=NULL;
  size=0;
  Add(v);
};

/*
ArrayOfVertices::~ArrayOfVertices() {               // destructor
  if (IsEmpty()) return;         // list is empty

  VertexList *pointer=anchor;
  while (pointer->next) pointer=pointer->next;         // set pointer to end

  do {
    pointer=pointer->last;              // go backwards through the list
    delete (pointer->next);            // delete the elements ...
  } while (pointer->last);              // until you are at the anchor
  
  delete (anchor);                   // delete anchor
};
*/

Vertex& ArrayOfVertices::operator[](size_t index) {
  return(GetItem(index));
};

void ArrayOfVertices::Add(Vertex v) {                         // add a vertex to the list
  if (IsEmpty()) {              // list is empty  
    anchor=new VertexList;
    anchor->last=NULL;
    anchor->next=NULL;
    anchor->element=v;  
  }
  else {
    VertexList *pointer=anchor; 
    while (pointer->next) pointer=pointer->next;        // set pointer to end
    pointer->next=new VertexList;
    (pointer->next)->last=pointer;
    pointer=pointer->next;
    pointer->next=NULL;
    pointer->element=v;
  }
  size++;
};

size_t ArrayOfVertices::Count(void) {                // count the elements in the array
  if (IsEmpty()) size=0;         // list is empty
  else {
    VertexList *pointer=anchor;    // pointer for going through the list
    size=1;                        // counter
    while (pointer->next) {  pointer=pointer->next; size++; }   // count elements
  }  
  return(size);
};

void ArrayOfVertices::Delete(size_t index) {                  // delete vertex at position index
  if (index<=0 || index>size) return;          // index out of boundaries

  VertexList *pointer=anchor;  // set pointer to start
  
  if(index==1) {      // anchor should be deleted
    if (size==1) anchor=NULL;  // delete anchor
    else {
      anchor=pointer->next;   // set second element as anchor
      anchor->last=NULL;
    }
  }
  else if (index==size) {    // last element should be deleted
    while (pointer->next) pointer=pointer->next;        // set pointer to end
    (pointer->last)->next=NULL;
  }
  else {
    for (unsigned int i=1; i<index; i++) pointer=pointer->next;         // go to right element
    (pointer->last)->next=pointer->next;
    (pointer->next)->last=pointer->last;
  }
  delete (pointer);                   // delete element
  size--;
};

void ArrayOfVertices::Empty(void) {                           // clears list
  if (IsEmpty()) return;              // list is empty

  VertexList *pointer=anchor;         // set pointer to end
  while (pointer->next) pointer=pointer->next;        // set pointer to end
  do {
    pointer=pointer->last;              // go backwards trhough the list
    delete (pointer->next);           // delete the elements ...
  } while (pointer->last);              // until you are at the anchor
  
  delete (anchor);                   // delete anchor
  anchor=NULL;
  size=0;
};

VertexList *ArrayOfVertices::GetAnchor(void) const {          // get anchor of the array
  return(anchor);
};

size_t ArrayOfVertices::GetCount(void) const {                // get size of the array
  return(size);
};

Vertex &ArrayOfVertices::GetItem(size_t index) {         // get item at position index
  if (index<=0 || index>size) {          // index out of boundaries
    if (!IsEmpty()) return(anchor->element);   // return first element, if there is one (else undefined)
  }  
  VertexList *pointer=anchor;           // set pointer to start
  for (unsigned int i=1; i<index; i++) pointer=pointer->next;         // go to right element
  return(pointer->element);                     // returns value
};

bool ArrayOfVertices::IsEmpty(void) const {                   // check if list is empty
  if (anchor) return(false);
  else return(true);
};

void ArrayOfVertices::SetItem(size_t index, Vertex v) {       // set item at position index to vertex
  if (index>0 && index<=size) {         // index out of boundaries
    VertexList *pointer=anchor;           // set pointer to start
    for (unsigned int i=1; i<index; i++) pointer=pointer->next;         // go to right element
    pointer->element=v;
  }  
};

// **************************
// Model Functions
// **************************

Model::Model(ArrayOfVertices vert) {          // standard contructor
  vertices=vert;
};

/*
Model::~Model() {               // destructor
  puts("M1");
  vertices.Empty();
  puts("M2");
};
*/

void Model::AddVertex(Vertex v) {                         // add a vertex to the list
  vertices.Add(v);
};

void Model::DeleteVertex(size_t index) {                  // delete vertex at position index
  vertices.Delete(index);
};

void Model::EmptyVertices(void) {                           // clears list
  vertices.Empty();
};

VertexList *Model::GetVertexAnchor(void) const {          // get anchor of the array
  return(vertices.GetAnchor());
};

size_t Model::GetVertexCount(void) const {                // get size of the array
  return(vertices.GetCount());
};

Vertex &Model::GetVertexItem(size_t index) {         // get item at position index
  return(vertices.GetItem(index));
};

bool Model::IsVertexEmpty(void) const {                   // check if list is empty
  return(vertices.IsEmpty());
};

void Model::SetVertexItem(size_t index, Vertex v) {       // set item at position index to vertex
  vertices.SetItem(index, v);
};

Model Model::Translate(Vector v) {             // move a Model
  for (unsigned int i=1; i<=vertices.GetCount(); i++) {
    vertices.SetItem(i, vertices[i].Translate(v));
  }
  return(*this);
};

Model Model::Rotate(double angle, Vector r) {    // rotate the Model around vector axis r
  for (unsigned int i=1; i<=vertices.GetCount(); i++) {
    vertices.SetItem(i, vertices[i].Rotate(angle,r));
  }
  return(*this);
};

Model Model::CreateQuader(double x, double y, double z) { // creates quader
  Vertex p1=Vertex(Vector(0,0,0));
  p1.AddEdge(Vector(x,0,0));
  p1.AddEdge(Vector(0,y,0));
  p1.AddEdge(Vector(0,0,z)); 
  
  Vertex p2=Vertex(Vector(x,0,z));
  p2.AddEdge(Vector(x,0,0));
  p2.AddEdge(Vector(x,y,z));
  p2.AddEdge(Vector(0,0,z)); 
  
  Vertex p3=Vertex(Vector(x,y,0));
  p3.AddEdge(Vector(x,0,0));
  p3.AddEdge(Vector(x,y,z));
  p3.AddEdge(Vector(0,y,0)); 
  
  Vertex p4=Vertex(Vector(0,y,z));
  p4.AddEdge(Vector(0,0,z));
  p4.AddEdge(Vector(x,y,z));
  p4.AddEdge(Vector(0,y,0)); 

  Model m;
  m.AddVertex(p1);
  m.AddVertex(p2);  
  m.AddVertex(p3);  
  m.AddVertex(p4);  
  return(m);
};

Model Model::CreateQuader(Vector v) {            // creates quader
  return(CreateQuader(v.x(), v.y(), v.z()));
};


// **************************
// ArrayOfModels Functions
// **************************

ArrayOfModels::ArrayOfModels(void) {        // standard constructor
  anchor=NULL;
  size=0;
};

ArrayOfModels::ArrayOfModels(Model v) {        // standard constructor
  anchor=NULL;
  size=0;
  Add(v);
};

/*
ArrayOfModels::~ArrayOfModels() {               // destructor
  if (IsEmpty()) return;         // list is empty

  ModelList *pointer=anchor;
  while (pointer->next) pointer=pointer->next;         // set pointer to end

  do {
    pointer=pointer->last;              // go backwards through the list
    delete (pointer->next);            // delete the elements ...
  } while (pointer->last);              // until you are at the anchor
  
  delete (anchor);                   // delete anchor
};
*/

Model& ArrayOfModels::operator[](size_t index) {
  return(GetItem(index));
};

void ArrayOfModels::Add(Model v) {                         // add a vertex to the list
  if (IsEmpty()) {              // list is empty  
    anchor=new ModelList;
    anchor->last=NULL;
    anchor->next=NULL;
    anchor->element=v;  
  }
  else {
    ModelList *pointer=anchor; 
    while (pointer->next) pointer=pointer->next;        // set pointer to end
    pointer->next=new ModelList;
    (pointer->next)->last=pointer;
    pointer=pointer->next;
    pointer->next=NULL;
    pointer->element=v;
  }
  size++;
};

size_t ArrayOfModels::Count(void) {                // count the elements in the array
  if (IsEmpty()) size=0;         // list is empty
  else {
    ModelList *pointer=anchor;    // pointer for going through the list
    size=1;                        // counter
    while (pointer->next) {  pointer=pointer->next; size++; }   // count elements
  }  
  return(size);
};

void ArrayOfModels::Delete(size_t index) {                  // delete vertex at position index
  if (index<=0 || index>size) return;          // index out of boundaries

  ModelList *pointer=anchor;  // set pointer to start
  
  if(index==1) {      // anchor should be deleted
    if (size==1) anchor=NULL;  // delete anchor
    else {
      anchor=pointer->next;   // set second element as anchor
      anchor->last=NULL;
    }
  }
  else if (index==size) {    // last element should be deleted
    while (pointer->next) pointer=pointer->next;        // set pointer to end
    (pointer->last)->next=NULL;
  }
  else {
    for (unsigned int i=1; i<index; i++) pointer=pointer->next;         // go to right element
    (pointer->last)->next=pointer->next;
    (pointer->next)->last=pointer->last;
  }
  delete (pointer);                   // delete element
  size--;
};

void ArrayOfModels::Empty(void) {                           // clears list
  if (IsEmpty()) return;              // list is empty

  ModelList *pointer=anchor;         // set pointer to end
  while (pointer->next) pointer=pointer->next;        // set pointer to end
  do {
    pointer=pointer->last;              // go backwards trhough the list
    delete (pointer->next);           // delete the elements ...
  } while (pointer->last);              // until you are at the anchor
  
  delete (anchor);                   // delete anchor
  anchor=NULL;
  size=0;
};

ModelList *ArrayOfModels::GetAnchor(void) const {          // get anchor of the array
  return(anchor);
};

size_t ArrayOfModels::GetCount(void) const {                // get size of the array
  return(size);
};

Model &ArrayOfModels::GetItem(size_t index) {         // get item at position index
  if (index<=0 || index>size) {          // index out of boundaries
    if (!IsEmpty()) return(anchor->element);   // return first element, if there is one (else undefined)
  }  
  ModelList *pointer=anchor;           // set pointer to start
  for (unsigned int i=1; i<index; i++) pointer=pointer->next;         // go to right element
  return(pointer->element);                     // returns value
};

bool ArrayOfModels::IsEmpty(void) const {                   // check if list is empty
  if (anchor) return(false);
  else return(true);
};

void ArrayOfModels::SetItem(size_t index, Model v) {       // set item at position index to vertex
  if (index>0 && index<=size) {         // index in boundaries
    ModelList *pointer=anchor;           // set pointer to start
    for (unsigned int i=1; i<index; i++) pointer=pointer->next;         // go to right element
    pointer->element=v;
  }  
};

// **************************
// Frame Functions
// **************************

ImageFrame::ImageFrame(const wxString& title, const wxPoint& pos, const wxSize& size, long style)      // standard constructor
   : wxFrame((wxFrame*) NULL, wxID_ANY, title, pos, size, style) {
                        
  Ex=Vector(1,0,0);                            // vector (1,0,0)
  Ey=Vector(0,1,0);                            // vector (0,1,0)
  Ez=Vector(0,0,1);                            // vector (0,0,1)

  ShowFullScreen(true);
  GetSize(&maxx, &maxy);     // maximum pixel that can be shown on screen
  WarpPointer(maxx/2, maxy/2);
  mousepos=wxPoint(maxx/2, maxy/2);
};

void ImageFrame::SetStandPoint(Vector v) {     // set new standpoint
  standpoint=v;		// position
};

void ImageFrame::SetViewPoint(Vector v) {     // set new viewpoint
  if (v.x()==0 && v.z()==0) return;  // if you are looking straight in the "sky" the program quits

  viewpoint=v;                // new view direction
  viewpoint.Normalize();     // normalize view vector
   
  /****************************************************
  The following calculations are some tricky.
  normvector2:
  - take the viewpoint vector and flat it to the x/z-plane
  - rotate it with 90 deg. around the y-axis (so it is necessary that x!=0 and z!=0 - if-statement above)
  - normalize it
  normvector1:
  - calculate some vector that is perpendicular to normvector 2 above and the viewpoint
  ****************************************************/
   
  normvector2=Vector(viewpoint.x(), 0, viewpoint.z());  //  normvector2 = sidewards vector to viewpoint in x-z plane
  normvector2.Rotate(90, Ey).Normalize();
  normvector1=viewpoint.Orth(normvector2);   // normvector1 = upwards orthogonal vector to viewpoint

/*  viewpoint.Print("viewpoint");
  normvector2.Print("normvector2");
  normvector1.Print("normvector1");  */
};

bool ImageFrame::IsVisible(Vector w, wxRealPoint *p) const {	// returns true if the point is visible in the frame

  Vector q=w-Vector(standpoint.x(), standpoint.y(), standpoint.z());    // current view in frame
  double l=q.Norm();
  double alpha, beta, gamma;
  Vector d;
  double a, b;

  if (l<0.1) {     // If you are standing to close to a point it should not be rendered.
    if (p) *p=wxRealPoint(0,0);
    return(false);
  }  

  beta=normvector1.Angle(q)-90;           // angle between normvector 1 and q with sign

  if (fabs(beta)>45) {     // The vertical angle is bigger than 45 deg. - the point is not visible.
    if (p) *p=wxRealPoint(0,0);
    return(false);
  }
  
  d=q+sin(PI*beta/180)*l*normvector1;                       // intersection point with view plane
  alpha=d.Angle(normvector2)-90;                      // angle between d and normvector2 with sign
  gamma=d.Angle(viewpoint);                          // angle between d and viewpoint without sign

  if (gamma>60) {              // The horizontal angle is bigger than 60 deg.
    if (p) *p=wxRealPoint(0,0);
    return(false);
  }

  a=-tan(PI*alpha/180);
  b=-tan(PI*beta/180)*sqrt(1+pow(a,2));
  if (p) *p=wxRealPoint( (1-a*0.75)*maxx/2, (1-b)*maxy/2 );
  return(true);
};

void ImageFrame::AddModel(Model v) {                         // add a vertex to the list
  scene.Add(v);
};

void ImageFrame::DeleteModel(size_t index) {                  // delete vertex at position index
  scene.Delete(index);
};

void ImageFrame::EmptyModels(void) {                           // clears list
  scene.Empty();
};

ModelList *ImageFrame::GetModelAnchor(void) const {          // get anchor of the array
  return(scene.GetAnchor());
};

size_t ImageFrame::GetModelCount(void) const {                // get size of the array
  return(scene.GetCount());
};

Model &ImageFrame::GetModelItem(size_t index) {         // get item at position index
  return(scene.GetItem(index));
};

bool ImageFrame::IsModelEmpty(void) const {                   // check if list is empty
  return(scene.IsEmpty());
};

void ImageFrame::SetModelItem(size_t index, Model v) {       // set item at position index to vertex
  scene.SetItem(index, v);
};

void ImageFrame::DrawLine(wxPaintDC *dc, Vector v, Vector w) {		// draw line between two points
  wxRealPoint p1, p2;
  double alpha;                  // angle between directions
  Vector v2=v, w2=w;
  Vector direc1, direc2;

  if (IsVisible(v, &p1) && IsVisible(w, &p2)) {  // both points are visible
    dc->DrawLine((wxCoord)round(p1.x), (wxCoord)round(p1.y),(wxCoord)round(p2.x), (wxCoord)round(p2.y));
    return;
  }  

  // One or both points are not on screen but there may be visible points in between.
  // calculate first both visible points starting from v and w

  if (v==w) {        // v==w, so there is only one point
    DrawPoint(dc, v);
    return;
  }
  // else v != w
  
  direc1=w-v2;   // calculate direction
  direc2=direc1;
  alpha=direc1.Angle(direc2);  // Angle between old and new direction - either 0 or 180 degree (here it is 0 of course)

  /*
  w.Print("w");
  v2.Print("v2");
  direc1.Print("direc1");
  direc2.Print("direc2");
  message(alpha, "alpha");
  */

  while (fabs(alpha)<1 && !IsVisible(v2)) {  // is alpha < 45 (so 0) you have not reached w 
    direc2.Normalize((v2-standpoint).Norm()/100);
    
    // Two points far away don't need a fine step width.

    if (direc2.Norm()>0.001) direc1=direc2;  // direction is not allowed to be too small

    v2+=direc1;    // go from v2 to w2    
    direc2=w-v2;    // calculate direction
    alpha=direc1.Angle(direc2);  // Angle between old and new direction - either 0 or 180 degree
  }

  // Is there a point between v and w that is visible, there is a second point calculated from w.
  if (IsVisible(v2)) {

    direc1=v-w2;   // calculate direction
    direc2=direc1;
    alpha=direc1.Angle(direc2);  // Angle between old and new direction - either 0 or 180 degree (here it is 0 of course)

    while (fabs(alpha)<1 && !IsVisible(w2)) {  // is alpha < 45 (so 0) you have not reached v 
      direc2.Normalize((w2-standpoint).Norm()/100);
      // Two points far away don't need a fine step width.

      if (direc2.Norm()>0.001) direc1=direc2;  // direction is not allowed to be too small

      w2+=direc1;    // go from v2 to w2    
      direc2=v-w2;    // calculate direction

      alpha=direc1.Angle(direc2);  // Angle between old and new direction - either 0 or 180 degree
    }
  }

  if (IsVisible(v2, &p1) && IsVisible(w2, &p2)) {  // both points are visible
    dc->DrawLine((wxCoord)round(p1.x), (wxCoord)round(p1.y),(wxCoord)round(p2.x), (wxCoord)round(p2.y));
    return;
  }
};

void ImageFrame::DrawPoint(wxPaintDC *dc, Vector v, wxCoord radius) {                            // draw a single point in space
  wxRealPoint p;
  if (IsVisible(v, &p)) {
    dc->DrawPoint((wxCoord)round(p.x), (wxCoord)round(p.y));
    if (radius)
      dc->DrawCircle((wxCoord)round(p.x), (wxCoord)round(p.y), radius);
  }    
};

void ImageFrame::DrawVector(wxPaintDC *dc, Vector v) {                           // draw the vector
  DrawLine(dc, v, Vector());
};

void ImageFrame::DrawVertex(wxPaintDC *dc, Vertex v) {                           // draw a vertex with all edges
  wxRealPoint p1, p2;
  Vector v1=v.GetVector();

  DrawPoint(dc, v1);
  VectorList *e=v.GetEdgeAnchor();        // set to anchor
  while (e) {                        // go through list
    DrawLine(dc, v1, e->element);
    e=e->next;          
  }
};

void ImageFrame::DrawModel(wxPaintDC *dc, Model m) {                           // draw the whole model
  VertexList *e=m.GetVertexAnchor();        // set to anchor
  while (e) {                              // go through list
    DrawVertex(dc, e->element);
    e=e->next;          
  }
};

void ImageFrame::DrawScene(wxPaintDC *dc) {                     // draw current scene on screen
  ModelList *e=scene.GetAnchor();         // set to anchor
  while (e) {                              // go through list
    DrawModel(dc, e->element);
    e=e->next;          
  }
};

void ImageFrame::OnPaint(wxPaintEvent& event) {   // paint wireframe
  wxPaintDC dc(this);                               // create device context             
  dc.BeginDrawing();                                
  DrawScene(&dc);
  dc.EndDrawing(); 
};

void ImageFrame::Look(wxMouseEvent & event) {   // mouse is moved
  wxPoint p=event.GetPosition();            // get mouse position
  Vector v=viewpoint, w=Vector(v.x(), 0, v.z());
  double alpha=viewpoint.Angle(w)*PI/180, beta=PI/180;
  double value, leftborder=0, rightborder=0, function=sin(alpha+beta)*viewpoint.Norm()-sin(alpha);
  
  if (p!=mousepos) {
    p=wxPoint(mousepos.x-p.x, mousepos.y-p.y);  // difference between old and new mouse position
    SetViewPoint(v.Rotate(p.x, Ey));     // look left / right
    v=viewpoint;    // get new viewpoint
    
    /* Calculate new look (up/down) if the angle is not to steep.
    I used the bisection method so that the angle of looking up and down is the same in each move. */
    
    if ((p.y>0 && Ey.Angle(viewpoint)>5) || (p.y<0 && Ey.Angle(viewpoint)<175)) {

      do { rightborder+=0.1; } while (function-rightborder>0);	// Liefert einen groben Schtzwert fr die zweite Grenze b.

      do {      // Calculate only zero with bisection method
        value=(leftborder+rightborder)/2.0;
        if ((function-leftborder)*(function-value)>0) leftborder=value;
        else rightborder=value;
      } while (function-value>0.000001);

      SetViewPoint(Vector(v.x(), v.y()+value*p.y, v.z()));     // look left / right
    }
    WarpPointer(maxx/2, maxy/2);
    Refresh();
  }

  if (event.LeftIsDown()) {
    SetStandPoint(standpoint+0.2*viewpoint);
    Refresh();
  }
  if (event.RightIsDown()) {
    SetStandPoint(standpoint-0.2*viewpoint);
    Refresh();
  }
 
};
